# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import os
import warnings

from nvshmem.core.nvshmem_types import NvshmemWarning

if os.path.exists(os.path.join(os.path.dirname(__file__), "rma.py")):
    from .rma import *
    from .direct import *
    from .amo import *
    from .collective import *
    from .mem import *
    __all__ = rma.__all__ + direct.__all__ + amo.__all__ + collective.__all__ + mem.__all__
else:
    warnings.warn("Numba device bindings are not enabled", NvshmemWarning)
    rma = None
    direct = None
    amo = None



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
The following are interoperability helpers for NVSHMEM4Py memory used with CuTe DSL

NOTE! CuTe DSL Tensor API aliases names with Torch tensors. Because of this, we do not import the cute module to the top level of the nvshmem.core module. 

Users of the CuTe DSL Tensor API should import the cute module directly via ``import nvshmem.core.interop.cute as cute``.
"""
import math
import ctypes
from enum import IntEnum
import nvshmem.core
from nvshmem.core.memory import buffer
from nvshmem.core.utils import get_size
from nvshmem.core._internal_tracking import _mr_references
from nvshmem.core.nvshmem_types import *
from nvshmem import bindings

from cuda.core import Buffer
from cuda.core import Device
from cuda.core import Stream

from typing import Any, Tuple, Union

__all__ = [
    "bytetensor", "tensor", "free_tensor", "tensor_get_buffer", "get_peer_tensor", "get_multicast_tensor",
    "register_external_tensor", "unregister_external_tensor", "cleanup_cute", "cute_compile_helper"
]

try:
    from cutlass import cute
    from cutlass.cute.typing import Pointer, Boolean, Int32, Int64, Constexpr, Float32, Int8
    from cutlass.cute import Tensor
    from cutlass.cute.runtime import from_dlpack
    _cute_enabled = True
except Exception as e:
    print(f"Error importing cutlass.cute: {e}")
    Tensor = None
    Pointer = None
    Boolean = None
    Int32 = None
    Int64 = None
    Constexpr = None
    Float32 = None
    Int8 = None
    _cute_enabled = False

_CUTE_TENSOR_BUFFERS = {}
_CUTE_MLIR_CONTEXT = None
_CUTE_MLIR_LOCATION = None
_CUTE_MLIR_MODULE = None

try:
    from cutlass._mlir import ir
    _cute_mlir_available = True
except Exception:
    ir = None
    _cute_mlir_available = False

_CUTE_DTYPE_NBYTES = {}
if _cute_enabled:
    _CUTE_DTYPE_NBYTES = {
        cute.Float16: 2,
        cute.BFloat16: 2,
        cute.Float32: 4,
        cute.Float64: 8,
        cute.Int8: 1,
        cute.Int16: 2,
        cute.Int32: 4,
        cute.Int64: 8,
        cute.Uint8: 1,
        cute.Uint16: 2,
        cute.Uint32: 4,
        cute.Uint64: 8,
        cute.Boolean: 1,
    }

# DLPack helpers for contiguous layout
try:
    import cuda.core._dlpack as _cuda_dlpack
    _dlpack_available = True
except Exception:
    _cuda_dlpack = None
    _dlpack_available = False


class _DLDevice(ctypes.Structure):
    _fields_ = [("device_type", ctypes.c_int), ("device_id", ctypes.c_int)]


class _DLDataType(ctypes.Structure):
    _fields_ = [("code", ctypes.c_uint8), ("bits", ctypes.c_uint8), ("lanes", ctypes.c_uint16)]


class _DLTensor(ctypes.Structure):
    _fields_ = [
        ("data", ctypes.c_void_p),
        ("device", _DLDevice),
        ("ndim", ctypes.c_int),
        ("dtype", _DLDataType),
        ("shape", ctypes.POINTER(ctypes.c_int64)),
        ("strides", ctypes.POINTER(ctypes.c_int64)),
        ("byte_offset", ctypes.c_uint64),
    ]


class _DLManagedTensor(ctypes.Structure):
    pass


_DLManagedTensorDeleter = ctypes.CFUNCTYPE(None, ctypes.POINTER(_DLManagedTensor))

_DLManagedTensor._fields_ = [
    ("dl_tensor", _DLTensor),
    ("manager_ctx", ctypes.c_void_p),
    ("deleter", _DLManagedTensorDeleter),
]

_DLPACK_TENSOR_HOLDERS = {}


def _safe_get_attr(obj, name, default=None):
    try:
        return getattr(obj, name)
    except Exception:
        return default


# Map CuTe dtype to DLPack code/bits
# https://github.com/dmlc/dlpack/blob/main/include/dlpack/dlpack.h
class _DLPackTypeCode(IntEnum):
    INT = 0
    UINT = 1
    FLOAT = 2
    BFLOAT = 4
    BOOL = 6


_CUTE_DLPACK_DTYPE = {
    cute.Float16: (_DLPackTypeCode.FLOAT, 16),
    cute.BFloat16: (_DLPackTypeCode.BFLOAT, 16),
    cute.Float32: (_DLPackTypeCode.FLOAT, 32),
    cute.Float64: (_DLPackTypeCode.FLOAT, 64),
    cute.Int8: (_DLPackTypeCode.INT, 8),
    cute.Int16: (_DLPackTypeCode.INT, 16),
    cute.Int32: (_DLPackTypeCode.INT, 32),
    cute.Int64: (_DLPackTypeCode.INT, 64),
    cute.Uint8: (_DLPackTypeCode.UINT, 8),
    cute.Uint16: (_DLPackTypeCode.UINT, 16),
    cute.Uint32: (_DLPackTypeCode.UINT, 32),
    cute.Uint64: (_DLPackTypeCode.UINT, 64),
    cute.Boolean: (_DLPackTypeCode.BOOL, 1),
}


@_DLManagedTensorDeleter
def _dlpack_deleter(ptr):
    if not ptr:
        return
    key = ctypes.addressof(ptr.contents)
    _DLPACK_TENSOR_HOLDERS.pop(key, None)


class _DLPackTensorWrapper:

    def __init__(self, capsule):
        self._capsule = capsule

    def __dlpack__(self, stream=None, max_version=None, dl_device=None, copy=None):
        return self._capsule

    def __dlpack_device__(self):
        # Not used by from_dlpack(), but provided for completeness
        return self._dlpack_device


def _make_dlpack_capsule(buf, shape, dtype, contiguous=True):
    if not _dlpack_available:
        raise NvshmemInvalid("cuda.core._dlpack is unavailable")
    if dtype not in _CUTE_DLPACK_DTYPE:
        raise NvshmemInvalid(f"Unsupported CuTe dtype for DLPack: {dtype}")
    # device info
    dev_type, dev_id = buf.__dlpack_device__()
    device = _DLDevice(int(dev_type), int(dev_id))
    code, bits = _CUTE_DLPACK_DTYPE[dtype]
    dl_dtype = _DLDataType(code=code, bits=bits, lanes=1)

    shape_arr = (ctypes.c_int64 * len(shape))(*shape)
    strides_ptr = None
    if not contiguous:
        raise NvshmemInvalid("Only contiguous layouts supported in DLPack path")

    dl_tensor = _DLTensor(
        data=ctypes.c_void_p(int(buf.handle)),
        device=device,
        ndim=len(shape),
        dtype=dl_dtype,
        shape=ctypes.cast(shape_arr, ctypes.POINTER(ctypes.c_int64)),
        strides=strides_ptr,
        byte_offset=0,
    )

    managed = _DLManagedTensor(
        dl_tensor=dl_tensor,
        manager_ctx=None,
        deleter=_dlpack_deleter,
    )
    key = ctypes.addressof(managed)
    _DLPACK_TENSOR_HOLDERS[key] = (managed, shape_arr)

    py_capsule = ctypes.pythonapi.PyCapsule_New
    py_capsule.restype = ctypes.py_object
    py_capsule.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_void_p]
    capsule = py_capsule(ctypes.addressof(managed), b"dltensor", None)
    wrapper = _DLPackTensorWrapper(capsule)
    wrapper._dlpack_device = (int(dev_type), int(dev_id))
    return wrapper


def _ensure_cute_mlir():
    global _CUTE_MLIR_CONTEXT, _CUTE_MLIR_LOCATION, _CUTE_MLIR_MODULE
    if not _cute_enabled:
        return
    if not _cute_mlir_available:
        raise NvshmemInvalid("CuTe MLIR context is not available")
    if _CUTE_MLIR_CONTEXT is None:
        _CUTE_MLIR_CONTEXT = ir.Context()
        _CUTE_MLIR_CONTEXT.__enter__()
        _CUTE_MLIR_LOCATION = ir.Location.unknown()
        _CUTE_MLIR_LOCATION.__enter__()
        _CUTE_MLIR_MODULE = ir.Module.create()


def _normalize_shape(shape):
    if isinstance(shape, int):
        return (shape, )
    return tuple(shape)


def _compute_strides(shape, morder):
    if len(shape) == 0:
        return ()
    if morder == "F":
        strides = [1]
        for dim in shape[:-1]:
            strides.append(strides[-1] * dim)
        return tuple(strides)
    strides = [1] * len(shape)
    running = 1
    for i in range(len(shape) - 1, -1, -1):
        strides[i] = running
        running *= shape[i]
    return tuple(strides)


def _size_in_bytes(shape, dtype):
    if dtype in _CUTE_DTYPE_NBYTES:
        return int(math.prod(shape) * _CUTE_DTYPE_NBYTES[dtype])
    try:
        return get_size(shape, dtype)
    except Exception as exc:
        raise NvshmemInvalid(f"Unsupported dtype for CuTe tensor: {dtype}") from exc


def _register_tensor_buffer(tensor, buf, dtype=None, shape=None, strides=None):
    _CUTE_TENSOR_BUFFERS[id(tensor)] = {
        "buffer": buf,
        "dtype": dtype,
        "element_type": dtype,  # Alias for consistency with tensor_get_buffer
        "shape": shape,
        "strides": strides,
    }


def _lookup_tensor_buffer(tensor):
    entry = _CUTE_TENSOR_BUFFERS.get(id(tensor))
    if entry is not None:
        return entry
    if isinstance(tensor, Tensor):
        return {
            "buffer": Buffer.from_handle(int(tensor.iterator), _size_in_bytes(tensor.shape, tensor.element_type)),
            "dtype": tensor.element_type,
            "shape": tensor.shape,
            "strides": tensor.stride,
        }
    return None


def _make_tensor_from_buffer(buf, shape, strides, dtype):
    _ensure_cute_mlir()
    with ir.InsertionPoint(_CUTE_MLIR_MODULE.body):
        ptr = cute.make_ptr(dtype, int(buf.handle), cute.AddressSpace.gmem)
        layout = cute.make_layout(tuple(shape), stride=tuple(strides))
        tensor = cute.make_tensor(ptr, layout)
    _register_tensor_buffer(tensor, buf, dtype=dtype, shape=shape, strides=strides)
    return tensor


def _is_tensor(tensor: Union[Tensor, object]) -> bool:
    """
    Helper function to check if an object is a CuTe DSL tensor
    This is used in collectives to avoid putting the complicated 
    import logic for CuTe DSL in any other file but this.
    """
    if not _cute_enabled:
        return False
    return isinstance(tensor, Tensor)


def tensor_get_buffer(tensor: Tensor) -> Tuple[Buffer, int, str]:
    """
    Get a nvshmem Buffer object from a CuTe DSL tensor object which was allocated with
    ``nvshmem.core.interop.cute.tensor()`` or ``nvshmem.core.interop.cute.bytetensor()``.
    Returns the buffer and the array's size.
    """
    entry = _lookup_tensor_buffer(tensor)
    if entry is None:
        raise NvshmemInvalid("Tried to retrieve buffer from Tensor not tracked by nvshmem")
    assert isinstance(tensor, Tensor), "Tried to register an external tensor that is not a CuTe DSL tensor"
    buf = entry.get("buffer")
    dtype = _safe_get_attr(tensor, "element_type")
    shape = entry.get("shape") or _safe_get_attr(tensor, "shape")
    size = _size_in_bytes(tuple(shape), dtype) if dtype is not None and shape is not None else buf.size
    return buf, size, dtype


def tensor(shape: Tuple[int], dtype: Any = Float32, release=False, morder="C", except_on_del=True) -> Tensor:
    """
    Create a CuTe tensor view on NVSHMEM-allocated memory with the given shape and dtype.

    This function allocates memory using NVSHMEM and constructs a CuTe tensor with the
    desired dtype, shape, and memory order. For contiguous layouts, it uses DLPack.
    """
    if not _cute_enabled:
        return

    if morder not in ("C", "F"):
        raise NvshmemInvalid("Tensor with invalid memory format requested")

    if dtype is None:
        dtype = Float32

    shape = _normalize_shape(shape)
    strides = _compute_strides(shape, morder)
    size = _size_in_bytes(shape, dtype)
    buf = buffer(size, release=release, except_on_del=except_on_del)

    if morder == "C" and _dlpack_available:
        capsule = _make_dlpack_capsule(buf, shape, dtype, contiguous=True)
        tensor = from_dlpack(capsule)
        _register_tensor_buffer(tensor, buf, dtype=dtype, shape=shape, strides=strides)
        return tensor

    return _make_tensor_from_buffer(buf, shape, strides, dtype)


def bytetensor(shape: Tuple[int], dtype: Any = Float32, release=False, morder="C", except_on_del=True) -> Tensor:
    """
    Create a CuTe tensor from NVSHMEM-allocated memory with the given shape and dtype.
    """
    if not _cute_enabled:
        return
    return tensor(shape, dtype=Int8, release=release, morder=morder, except_on_del=except_on_del)


def get_peer_tensor(tensor: Tensor, peer_pe: int = None) -> Tensor:
    """
    Return a Buffer based on the ``peer_buffer`` (wrapper of nvshmem_ptr) API
    """
    if not _cute_enabled:
        return
    buf, _, _ = tensor_get_buffer(tensor)
    peer_buf = nvshmem.core.get_peer_buffer(buf, peer_pe)
    return _make_tensor_from_buffer(peer_buf, tensor.shape, tensor.stride, tensor.element_type)


def get_multicast_tensor(team: Teams, tensor: Tensor) -> Tensor:
    """
    Returns a CuTe Tensor view on multicast-accessible memory corresponding to the input tensor.
    """
    if not _cute_enabled:
        return
    buf, _, _ = tensor_get_buffer(tensor)
    mc_buf = nvshmem.core.get_multicast_buffer(team, buf)
    return _make_tensor_from_buffer(mc_buf, tensor.shape, tensor.stride, tensor.element_type)


def register_external_tensor(tensor: Tensor) -> Tensor:
    """
    Register an external tensor with NVSHMEM.
    """
    if not _cute_enabled:
        return
    buf, _, _ = tensor_get_buffer(tensor)
    registered_buf = nvshmem.core.register_external_buffer(buf)
    return _make_tensor_from_buffer(registered_buf, tensor.shape, tensor.stride, tensor.element_type)


def unregister_external_tensor(tensor: Tensor) -> None:
    """
    Unregister an external tensor with NVSHMEM.
    """
    if not _cute_enabled:
        return
    buf, _, _ = tensor_get_buffer(tensor)
    nvshmem.core.unregister_external_buffer(buf)


def cleanup_cute():
    """
    Release CuTe MLIR context and DLPack holders to avoid teardown crashes.
    """
    global _CUTE_MLIR_CONTEXT, _CUTE_MLIR_LOCATION, _CUTE_MLIR_MODULE
    _DLPACK_TENSOR_HOLDERS.clear()
    try:
        if _CUTE_MLIR_LOCATION is not None:
            _CUTE_MLIR_LOCATION.__exit__(None, None, None)
    except Exception:
        pass
    try:
        if _CUTE_MLIR_CONTEXT is not None:
            _CUTE_MLIR_CONTEXT.__exit__(None, None, None)
    except Exception:
        pass
    _CUTE_MLIR_LOCATION = None
    _CUTE_MLIR_CONTEXT = None
    _CUTE_MLIR_MODULE = None


def free_tensor(tensor: Tensor) -> None:
    """
    Free an NVSHMEM-backed CuTe DSL Tensor

    Args:
        tensor (``cute.Tensor``): A CuTe tensor backed by NVSHMEM memory.

    Returns:
        ``None``

    Raises:
        ``NvshmemInvalid``: If the input tensor is not backed by NVSHMEM memory.
    """
    if not _cute_enabled:
        return
    # Convert array to Buffer
    buf, sz, dtype = tensor_get_buffer(tensor)
    nvshmem.core.free(buf)


def cute_compile_helper(kernel_fn, *args, **kwargs):
    """
    Helper function to compile a CuTe DSL kernel function.

    Finds the current device's NVSHMEM bitcode library and compiles the kernel function with it.

    Runs nvshmem.core.library_init with the compiled kernel.

    Args:
        kernel_fn: A CuTe kernel function decorated with @cute.jit that contains a launcher.
                   The launcher should call a @cute.kernel with .launch().
        *args: Example arguments for compilation (tensors, etc.)
        **kwargs: Additional arguments passed to cute.compile()

    Returns:
        A tuple containing:
        - The compiled kernel function. (a callable object)
        - The nvshmem kernel object. (a NvshmemKernelObject) - the user should run ``nvshmem.core.library_finalize`` 
                  on this object after the kernel is executed.

    NOTE: This function assumes that the device being used as the NVSHMEM PE is already set current.
    """
    nvshmem_device_bc = nvshmem.core.find_device_bitcode_library()
    # Important: If _CUTE_MLIR_MODULE exists (from tensor creation via _make_tensor_from_buffer),
    # its context is active. cute.compile() checks "if ir.Context.current is None" and if not,
    # tries to access "ir.InsertionPoint.current" which raises an error if no insertion point is active.
    #
    # Solution: Temporarily exit our context and location if they're active, so cute.compile()
    # can create its own context. Then re-enter them after compilation.
    context_exited = False
    location_exited = False

    if _CUTE_MLIR_CONTEXT is not None:
        # Check if our context is currently active
        try:
            current_ctx = ir.Context.current
            if current_ctx is not None:
                # Our context might be active - exit it temporarily
                # We need to exit location first, then context
                if _CUTE_MLIR_LOCATION is not None:
                    _CUTE_MLIR_LOCATION.__exit__(None, None, None)
                    location_exited = True
                _CUTE_MLIR_CONTEXT.__exit__(None, None, None)
                context_exited = True
        except Exception:
            # If we can't check, assume context is not active
            pass

    try:
        # Build compile_kwargs with options and any user-provided kwargs
        compile_kwargs = {"options": f" --link-libraries={nvshmem_device_bc}"}
        if kwargs:
            compile_kwargs.update(kwargs)

        # Call cute.compile() - it will create its own context if needed
        compilerd_func = cute.compile(kernel_fn, *args, **compile_kwargs)
    finally:
        # Re-enter our context and location if we exited them
        if context_exited and _CUTE_MLIR_CONTEXT is not None:
            _CUTE_MLIR_CONTEXT.__enter__()
        if location_exited and _CUTE_MLIR_LOCATION is not None:
            _CUTE_MLIR_LOCATION.__enter__()
    # NOTE! assumes that device is already set current.
    dev = Device()
    compilerd_func = compilerd_func.to(dev.device_id)
    cuda_library = compilerd_func.jit_module.cuda_library
    nvshmem_kernel = nvshmem.core.NvshmemKernelObject.from_handle(int(cuda_library[0]))
    nvshmem.core.library_init(nvshmem_kernel)
    return compilerd_func, nvshmem_kernel




# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# SPDX-License-Identifier: Apache-2.0
#
# This code was automatically generated with version 3.3.0. Do not modify it directly.

from ..cynvshmem cimport *


###############################################################################
# Wrapper functions
###############################################################################

cdef int _nvshmem_barrier(nvshmem_team_t team) except* nogil
cdef void _nvshmem_barrier_all() except* nogil
cdef int _nvshmemx_init_status() except* nogil
cdef int _nvshmem_my_pe() except* nogil
cdef int _nvshmem_n_pes() except* nogil
cdef void _nvshmem_info_get_version(int* major, int* minor) except* nogil
cdef void _nvshmemx_vendor_get_version_info(int* major, int* minor, int* patch) except* nogil
cdef void* _nvshmem_malloc(size_t size) except* nogil
cdef void* _nvshmem_calloc(size_t count, size_t size) except* nogil
cdef void* _nvshmem_align(size_t count, size_t size) except* nogil
cdef void _nvshmem_free(void* ptr) except* nogil
cdef void* _nvshmem_ptr(const void* dest, int pe) except* nogil
cdef void* _nvshmemx_mc_ptr(nvshmem_team_t team, const void* ptr) except* nogil
cdef int _nvshmem_team_my_pe(nvshmem_team_t team) except* nogil
cdef int _nvshmem_team_n_pes(nvshmem_team_t team) except* nogil
cdef void _nvshmem_team_get_config(nvshmem_team_t team, nvshmem_team_config_t* config) except* nogil
cdef int _nvshmem_team_translate_pe(nvshmem_team_t src_team, int src_pe, nvshmem_team_t dest_team) except* nogil
cdef int _nvshmem_team_split_strided(nvshmem_team_t parent_team, int PE_start, int PE_stride, int PE_size, const nvshmem_team_config_t* config, long config_mask, nvshmem_team_t* new_team) except* nogil
cdef int _nvshmemx_team_get_uniqueid(nvshmemx_team_uniqueid_t* uniqueid) except* nogil
cdef int _nvshmemx_team_init(nvshmem_team_t* team, nvshmem_team_config_t* config, long config_mask, int npes, int pe_idx_in_team) except* nogil
cdef int _nvshmem_team_split_2d(nvshmem_team_t parent_team, int xrange, const nvshmem_team_config_t* xaxis_config, long xaxis_mask, nvshmem_team_t* xaxis_team, const nvshmem_team_config_t* yaxis_config, long yaxis_mask, nvshmem_team_t* yaxis_team) except* nogil
cdef void _nvshmem_team_destroy(nvshmem_team_t team) except* nogil
cdef int _nvshmemx_bfloat16_alltoall_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_half_alltoall_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_float_alltoall_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_double_alltoall_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_char_alltoall_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_short_alltoall_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_schar_alltoall_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int_alltoall_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_long_alltoall_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_longlong_alltoall_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int8_alltoall_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int16_alltoall_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int32_alltoall_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int64_alltoall_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint8_alltoall_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint16_alltoall_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint32_alltoall_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint64_alltoall_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_size_alltoall_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_barrier_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil
cdef void _nvshmemx_barrier_all_on_stream(cudaStream_t stream) except* nogil
cdef int _nvshmemx_team_sync_on_stream(nvshmem_team_t team, cudaStream_t stream) except* nogil
cdef void _nvshmemx_sync_all_on_stream(cudaStream_t stream) except* nogil
cdef int _nvshmemx_bfloat16_broadcast_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_half_broadcast_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_float_broadcast_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_double_broadcast_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_char_broadcast_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_short_broadcast_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_schar_broadcast_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int_broadcast_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_long_broadcast_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_longlong_broadcast_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int8_broadcast_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int16_broadcast_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int32_broadcast_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int64_broadcast_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint8_broadcast_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint16_broadcast_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint32_broadcast_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint64_broadcast_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_size_broadcast_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, int PE_root, cudaStream_t stream) except* nogil
cdef int _nvshmemx_bfloat16_fcollect_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_half_fcollect_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_float_fcollect_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_double_fcollect_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_char_fcollect_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_short_fcollect_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_schar_fcollect_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int_fcollect_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_long_fcollect_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_longlong_fcollect_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int8_fcollect_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int16_fcollect_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int32_fcollect_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int64_fcollect_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint8_fcollect_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint16_fcollect_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint32_fcollect_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint64_fcollect_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_size_fcollect_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nelem, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int8_max_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int16_max_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int32_max_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int64_max_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint8_max_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint16_max_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint32_max_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint64_max_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_size_max_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_char_max_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_schar_max_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_short_max_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int_max_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_long_max_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_longlong_max_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_bfloat16_max_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_half_max_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_float_max_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_double_max_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int8_min_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int16_min_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int32_min_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int64_min_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint8_min_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint16_min_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint32_min_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint64_min_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_size_min_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_char_min_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_schar_min_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_short_min_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int_min_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_long_min_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_longlong_min_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_bfloat16_min_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_half_min_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_float_min_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_double_min_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int8_sum_reduce_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int16_sum_reduce_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int32_sum_reduce_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int64_sum_reduce_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint8_sum_reduce_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint16_sum_reduce_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint32_sum_reduce_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint64_sum_reduce_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_size_sum_reduce_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_char_sum_reduce_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_schar_sum_reduce_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_short_sum_reduce_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int_sum_reduce_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_long_sum_reduce_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_longlong_sum_reduce_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_bfloat16_sum_reduce_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_half_sum_reduce_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_float_sum_reduce_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_double_sum_reduce_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int8_max_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int16_max_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int32_max_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int64_max_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint8_max_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint16_max_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint32_max_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint64_max_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_size_max_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_char_max_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_schar_max_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_short_max_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int_max_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_long_max_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_longlong_max_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_bfloat16_max_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_half_max_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_float_max_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_double_max_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int8_min_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int16_min_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int32_min_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int64_min_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint8_min_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint16_min_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint32_min_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint64_min_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_size_min_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_char_min_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_schar_min_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_short_min_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int_min_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_long_min_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_longlong_min_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_bfloat16_min_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_half_min_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_float_min_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_double_min_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int8_sum_reducescatter_on_stream(nvshmem_team_t team, int8_t* dest, const int8_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int16_sum_reducescatter_on_stream(nvshmem_team_t team, int16_t* dest, const int16_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int32_sum_reducescatter_on_stream(nvshmem_team_t team, int32_t* dest, const int32_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int64_sum_reducescatter_on_stream(nvshmem_team_t team, int64_t* dest, const int64_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint8_sum_reducescatter_on_stream(nvshmem_team_t team, uint8_t* dest, const uint8_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint16_sum_reducescatter_on_stream(nvshmem_team_t team, uint16_t* dest, const uint16_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint32_sum_reducescatter_on_stream(nvshmem_team_t team, uint32_t* dest, const uint32_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_uint64_sum_reducescatter_on_stream(nvshmem_team_t team, uint64_t* dest, const uint64_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_size_sum_reducescatter_on_stream(nvshmem_team_t team, size_t* dest, const size_t* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_char_sum_reducescatter_on_stream(nvshmem_team_t team, char* dest, const char* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_schar_sum_reducescatter_on_stream(nvshmem_team_t team, signed char* dest, const signed char* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_short_sum_reducescatter_on_stream(nvshmem_team_t team, short* dest, const short* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_int_sum_reducescatter_on_stream(nvshmem_team_t team, int* dest, const int* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_long_sum_reducescatter_on_stream(nvshmem_team_t team, long* dest, const long* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_longlong_sum_reducescatter_on_stream(nvshmem_team_t team, long long* dest, const long long* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_bfloat16_sum_reducescatter_on_stream(nvshmem_team_t team, __nv_bfloat16* dest, const __nv_bfloat16* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_half_sum_reducescatter_on_stream(nvshmem_team_t team, half* dest, const half* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_float_sum_reducescatter_on_stream(nvshmem_team_t team, float* dest, const float* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_double_sum_reducescatter_on_stream(nvshmem_team_t team, double* dest, const double* src, size_t nreduce, cudaStream_t stream) except* nogil
cdef int _nvshmemx_hostlib_init_attr(unsigned int flags, nvshmemx_init_attr_t* attr) except* nogil
cdef void _nvshmemx_hostlib_finalize() except* nogil
cdef int _nvshmemx_set_attr_uniqueid_args(const int myrank, const int nranks, const nvshmemx_uniqueid_t* uniqueid, nvshmemx_init_attr_t* attr) except* nogil
cdef int _nvshmemx_set_attr_mpi_comm_args(void* mpi_comm, nvshmemx_init_attr_t* nvshmem_attr) except* nogil
cdef int _nvshmemx_get_uniqueid(nvshmemx_uniqueid_t* uniqueid) except* nogil
cdef int _nvshmemx_cumodule_init(CUmodule module) except* nogil
cdef int _nvshmemx_cumodule_finalize(CUmodule module) except* nogil
cdef void* _nvshmemx_buffer_register_symmetric(void* buf_ptr, size_t size, int flags) except* nogil
cdef int _nvshmemx_buffer_unregister_symmetric(void* mmap_ptr, size_t size) except* nogil
cdef int _nvshmemx_culibrary_init(CUlibrary library) except* nogil
cdef int _nvshmemx_culibrary_finalize(CUlibrary library) except* nogil
cdef void _nvshmemx_putmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil
cdef void _nvshmemx_putmem_signal_on_stream(void* dest, const void* source, size_t bytes, uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil
cdef void _nvshmemx_getmem_on_stream(void* dest, const void* source, size_t bytes, int pe, cudaStream_t cstrm) except* nogil
cdef void _nvshmemx_quiet_on_stream(cudaStream_t cstrm) except* nogil
cdef void _nvshmemx_signal_op_on_stream(uint64_t* sig_addr, uint64_t signal, int sig_op, int pe, cudaStream_t cstrm) except* nogil
cdef void _nvshmemx_signal_wait_until_on_stream(uint64_t* sig_addr, int cmp, uint64_t cmp_value, cudaStream_t cstream) except* nogil



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
The following are NVSHMEM4Py APIs that expose host-initiated collective communication
"""

from cuda.core import Stream
from cuda.core import Buffer
from cuda.core import Device

from nvshmem.core.utils import _get_device, dtype_nbytes
from nvshmem.core.interop.cupy import _is_array, array_get_buffer
from nvshmem.core.interop.torch import _is_tensor, tensor_get_buffer
from nvshmem.core._internal_tracking import _mr_references, _is_initialized, InternalInitStatus
from nvshmem.core.direct import team_n_pes, n_pes
from nvshmem.core.nvshmem_types import *
import nvshmem.bindings as bindings

from typing import Union, Tuple
import logging

logger = logging.getLogger("nvshmem")

__all__ = [
    "reduce", "reducescatter", "alltoall", "fcollect", "broadcast", "barrier", "sync", "barrier_all", "sync_all",
    "collective_on_buffer"
]
"""
On-Stream Collectives

All take a team, a Stream. Some take an operation.
"""

# Valid reduction/reducescatter ops. Enhance if needed.
valid_ops = ["min", "max", "sum", None]

# Collectives
valid_collectives = ["reduce", "reducescatter", "alltoall", "fcollect", "broadcast"]

# Mapping from Cupy/Torch dtypes to NVSHMEM dtype names
external_to_nvshmem_dtypes = {
    # --------------------
    # Torch dtypes
    # --------------------
    "torch.float16": "half",
    "torch.bfloat16": "bfloat16",
    "torch.float32": "float",
    "torch.float64": "double",
    "torch.uint8": "uint8",
    "torch.int8": "int8",
    "torch.int16": "int16",
    "torch.int32": "int32",
    "torch.int64": "int64",
    "torch.bool": "uint8",  # mapped to uint8 for collective support

    # --------------------
    # CuPy dtypes
    # --------------------
    "float16": "half",
    "bfloat16": "bfloat16",
    "float32": "float",
    "float64": "double",
    "uint8": "uint8",
    "int8": "int8",
    "int16": "int16",
    "int32": "int32",
    "int64": "int64",
    "bool": "uint8",  # mapped to uint8 for collective support
}


def _call_collective(coll: str,
                     team: Teams,
                     dst_array: object,
                     src_array: object,
                     op: str = None,
                     root: int = 0,
                     stream: Stream = None):
    """
    Executes a collective operation on structured array types (Torch or CuPy).

    This function performs validation and dispatches the appropriate NVSHMEM
    collective routine using the binding layer. It is intended for host-initiated
    collectives operating on array types.

    Args:
        - team: The NVSHMEM team handle.
        - src_array (object): The source array (Torch Tensor or CuPy ndarray).
        - dst_array (object): The destination array (Torch Tensor or CuPy ndarray).
        - coll (str): Collective operation name (e.g., "reduce", "broadcast").
        - op (str, optional): Reduction operator (e.g., "sum", "max"). Defaults to None.
        - stream (Stream, optional): CUDA stream to associate with the operation. Defaults to None.
        - root: Root PE. Currently only used for broadcast. Defaults to 0

    Raises:
        NvshmemInvalid: If the input arrays are incompatible or an invalid operator is used.
    """

    # Excepts upon invalid dtype
    src_buf, src_buf_size, src_nvshmem_dtype = _check_dtype(src_array)
    dst_buf, dst_buf_size, dst_nvshmem_dtype = _check_dtype(dst_array)

    if src_nvshmem_dtype != dst_nvshmem_dtype:
        raise NvshmemInvalid("Non-matching data types for src and dest arrays")

    if op not in valid_ops:
        raise NvshmemInvalid(f"Invalid operator passed to reduce_on_stream. Valid options: {valid_ops.join(', ', )}")

    collective_on_buffer(coll, team, dst_buf, src_buf, dtype=src_nvshmem_dtype, op=op, root=root, stream=stream)


def collective_on_buffer(coll: str,
                         team: Teams,
                         dest: Buffer,
                         src: Buffer,
                         dtype: str = None,
                         op: str = None,
                         root: int = 0,
                         stream: Stream = None,
                         enable_timing=False) -> float:
    """
    This function allows host-initiated collectives over raw memory buffers.
    It is used by higher-level wrappers or when working with DLPack-converted memory directly.

    Args:
        - team: The NVSHMEM team handle.
        - src (Buffer): Source buffer.
        - dest (Buffer): Destination buffer.
            Source and dest buffers MUST be allocated by nvshmem4py
        - coll (str): Collective type (e.g., "reduce", "fcollect").
        - dtype (str, optional): NVSHMEM dtype string for function resolution.
        - op (str, optional): Reduction operator if required.
        - stream (Stream, optional): CUDA stream to execute on.
        - root (int): Root PE for collective. Only used for Broadcast today.
        - enable_timing: If True, return the time (in ms) it took the collective to execute

    Raises:
        ``NvshmemInvalid``: If the Buffer or collective type is invalid.

    Returns:
        float: time in ms that the collective took to execute if enable_timing is set, else 0

    NOTE:
        This API is considered experimental and should be used only by expert users
    """
    if _is_initialized["status"] != InternalInitStatus.INITIALIZED:
        raise NvshmemInvalid("NVSHMEM Library is not initialized")

    if stream is None:
        # TODO: Implement default stream and support non-on-stream collectives
        logger.warning("Non-on-stream collectives are not yet implemented. Stream may not be None")
        raise NotImplemented

    user_nvshmem_dev, other_dev = _get_device()
    if not isinstance(src, Buffer) or not isinstance(dest, Buffer):
        raise NvshmemInvalid("Called collective on an invalid Buffer")

    # Assert that these buffers are tracked by nvshmem
    if not _mr_references.get(user_nvshmem_dev.device_id, {})._mem_references.get(int(src.handle)) \
       or not _mr_references.get(user_nvshmem_dev.device_id, {})._mem_references.get(int(dest.handle)):
        raise NvshmemInvalid("Tried to perform a collective on a buffer not allocated by NVSHMEM4Py")

    if not coll in valid_collectives:
        raise NvshmemInvalid("Passed invalid collective")

    # Set up timing
    if enable_timing:
        start_event = user_nvshmem_dev.create_event({"enable_timing": True})
        stop_event = user_nvshmem_dev.create_event({"enable_timing": True})
    else:
        time_ms = 0

    # Op and dtype validity are checked in _call_collective()
    # For callers of this function directly, we don't know anymore what the types are
    # - it's on the user to get it right.

    src_size = src.size
    dest_size = dest.size
    if coll in ("fcollect", "alltoall"):
        size = src_size
    else:
        size = dest_size

    if coll != "alltoall":
        size_elem = max(1, size // dtype_nbytes(dtype))
    else:
        size_elem = max(1, size // (n_pes() * dtype_nbytes(dtype)))

    func_name = ""
    if dtype:
        func_name += f"{dtype}_"
    if op:
        func_name += f"{op}_"
    func_name += f"{coll}_on_stream"

    # We have a string of the coll function name
    coll_func = getattr(bindings, func_name)
    function_args = [team, int(dest.handle), int(src.handle), size_elem, int(stream.__cuda_stream__()[1])]
    if coll == "broadcast":
        # Rewrite the whole thing instead of append to make the ordering requirements more obvious
        function_args = [team, int(dest.handle), int(src.handle), size_elem, root, int(stream.__cuda_stream__()[1])]
    if enable_timing:
        stream.record(start_event)
    result = coll_func(*function_args)

    if enable_timing:
        stream.record(stop_event)
        stop_event.sync()
        time_ms = (stop_event - start_event)  # in ms

    # If result is not 0 or None, it failed
    if result:
        raise NvshmemError(f"Collective {coll} on team {team} failed to execute")

    if other_dev is not None:
        other_dev.set_current()

    return time_ms


def _check_dtype(array: object) -> Tuple[Buffer, int, str]:
    """
    Validates an array for collective operations and extracts metadata.

    This function ensures the array is a supported type (Torch or CuPy) and maps
    its dtype to an NVSHMEM-compatible type.

    Args:
        array (object): The array object to inspect.

    Returns:
        Tuple[Buffer, int, str]: A tuple containing the device buffer, the size in bytes,
                                 and the NVSHMEM dtype string.

    Raises:
        ``NvshmemInvalid``: If the array is not a supported type or has an unsupported dtype.
    """
    # The correct device must be current here because we need to know the device ID to look up the correct MR
    user_nvshmem_dev, other_dev = _get_device()
    is_array = _is_array(array)
    is_tensor = _is_tensor(array)
    if not (is_array or is_tensor):
        raise NvshmemInvalid("Passed a non-array object into a collective function")

    if is_array:
        buf, buf_size, external_dtype = array_get_buffer(array)

    elif is_tensor:
        buf, buf_size, external_dtype = tensor_get_buffer(array)

    nvshmem_dtype = external_to_nvshmem_dtypes.get(external_dtype)
    if not nvshmem_dtype:
        raise NvshmemInvalid("Passed an invalid datatype into a collective function")
    if other_dev is not None:
        other_dev.set_current()
    return buf, buf_size, nvshmem_dtype


def barrier(team: Teams, stream: NvshmemStreamsType = None) -> None:
    """
    Executes a team-wide barrier on a specified CUDA stream.

    Does not support MPG use cases

    Args:
        - team: NVSHMEM team handle.
        - stream (Stream): CUDA stream for synchronization.
    """
    if stream is None:
        raise NvshmemInvalid("Stream may not be None for barrier")
    user_nvshmem_dev, other_dev = _get_device()
    bindings.barrier_on_stream(team, int(stream.__cuda_stream__()[1]))
    if other_dev is not None:
        other_dev.set_current()


def sync(team: Teams, stream: NvshmemStreamsType = None) -> None:
    """
    Executes a team-wide sync on a specified CUDA stream.

    Does not support MPG use cases

    Args:
        - team: NVSHMEM team handle.
        - stream (Stream): CUDA stream for synchronization.
    """
    if stream is None:
        raise NvshmemInvalid("Stream may not be None for sync")
    user_nvshmem_dev, other_dev = _get_device()
    bindings.team_sync_on_stream(team, int(stream.__cuda_stream__()[1]))
    if other_dev is not None:
        other_dev.set_current()


def barrier_all(stream: NvshmemStreamsType = None) -> None:
    """
    Executes a runtime-wide barrier on a specified CUDA stream.

    Supports MPG use cases

    Args:
        - stream (Stream): CUDA stream for synchronization.
    """
    if stream is None:
        raise NvshmemInvalid("Stream may not be None for barrier_all")
    user_nvshmem_dev, other_dev = _get_device()
    bindings.barrier_all_on_stream(int(stream.__cuda_stream__()[1]))
    if other_dev is not None:
        other_dev.set_current()


def sync_all(stream: NvshmemStreamsType = None) -> None:
    """
    Executes a runtime-wide sync on a specified CUDA stream.

    Supports MPG use cases

    Args:
        - stream (Stream): CUDA stream for synchronization.
    """
    if stream is None:
        raise NvshmemInvalid("Stream may not be None for sync_all")
    user_nvshmem_dev, other_dev = _get_device()
    bindings.sync_all_on_stream(int(stream.__cuda_stream__()[1]))
    if other_dev is not None:
        other_dev.set_current()


def reduce(team: Teams, dst_array: object, src_array: object, op: str, stream: NvshmemStreamsType = None):
    """
    Performs a reduction from src_array to dst_array on a CUDA stream.

    Args:
        - src_array: Source array (Torch or CuPy).
        - dst_array: Destination array.
        - team: NVSHMEM team handle.
        - stream (Stream): CUDA stream to perform the operation on.
        - op (str): Reduction operator (e.g., "sum").
    """
    _call_collective("reduce", team, dst_array, src_array, op=op, stream=stream)


def reducescatter(team: Teams, dst_array: object, src_array: object, op: str, stream: NvshmemStreamsType = None):
    """
    Performs a reduce-scatter operation on a CUDA stream.

    Args:
        - src_array: Source array (Torch or CuPy).
        - dst_array: Destination array.
        - team: NVSHMEM team handle.
        - stream (Stream): CUDA stream to perform the operation on.
        - op (str): Reduction operator (e.g., "sum").
    """
    _call_collective("reducescatter", team, dst_array, src_array, op=op, stream=stream)


def alltoall(team: Teams, dst_array: object, src_array: object, stream: NvshmemStreamsType = None):
    """
    Performs an all-to-all communication on a CUDA stream.

    Args:
        - src_array: Source array (Torch or CuPy).
        - dst_array: Destination array.
        - team: NVSHMEM team handle.
        - stream (Stream): CUDA stream to perform the operation on.
    """
    _call_collective("alltoall", team, dst_array, src_array, op=None, stream=stream)


def fcollect(team: Teams, dst_array: object, src_array: object, stream: NvshmemStreamsType = None):
    """
    Performs a full-collective operation on a CUDA stream.

    Args:
        - src_array: Source array (Torch or CuPy).
        - dst_array: Destination array.
        - team: NVSHMEM team handle.
        - stream (Stream): CUDA stream to perform the operation on.
    """
    _call_collective("fcollect", team, dst_array, src_array, op=None, stream=stream)


def broadcast(team: Teams, dst_array: object, src_array: object, root: int = 0, stream: NvshmemStreamsType = None):
    """
    Broadcasts data from src_array to dst_array across the team on a CUDA stream.

    Args:
        - src_array: Source array (Torch or CuPy).
        - dst_array: Destination array.
        - team: NVSHMEM team handle.
        - stream (Stream): CUDA stream to perform the operation on.
        - root: Root PE
    """
    _call_collective("broadcast", team, dst_array, src_array, root=root, op=None, stream=stream)



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from nvshmem.bindings.device.cute import *

import cutlass
from cutlass import cute
from cutlass.base_dsl.ast_helpers import const_expr
from cutlass.base_dsl.typing import cast as cute_cast

__all__ = [
    "atomic_inc",
    "atomic_fetch_inc",
    "atomic_fetch",
    "atomic_set",
    "atomic_add",
    "atomic_fetch_add",
    "atomic_and",
    "atomic_fetch_and",
    "atomic_or",
    "atomic_fetch_or",
    "atomic_xor",
    "atomic_fetch_xor",
    "atomic_swap",
    "atomic_compare_swap",
]


@cute.jit
def _resolve_ptr(arg):
    return arg.iterator


@cute.jit
def _resolve_dtype(dst):
    return dst.element_type


@cute.jit
def atomic_fetch(src, pe):
    """
    Atomically fetches (reads) the current value at symmetric ``src`` on PE ``pe``.

    This is a thread-level remote atomic operation. The read is performed atomically
    with respect to other atomic operations on the same location.

    Args:
        - ``src``: CuTe tensor view pointing to a single-element symmetric source on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. The element dtype determines
          which underlying NVSHMEM atomic is dispatched.
        - ``pe`` (``int``): Source PE to fetch from.

    Returns:
        The current value stored at ``src`` on PE ``pe``, with the same dtype as ``src``.

    Note:
        Supported dtypes are determined by the NVSHMEM atomic fetch dispatch table.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    src_ptr = _resolve_ptr(src)
    dtype = _resolve_dtype(src)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_fetch(src_ptr, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_fetch(src_ptr, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_fetch(src_ptr, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_fetch(src_ptr, pe)


    elif const_expr(dtype == cutlass.Float32):
        return float_atomic_fetch(src_ptr, pe)


    elif const_expr(dtype == cutlass.Float64):
        return double_atomic_fetch(src_ptr, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_fetch: {dtype}")


@cute.jit
def atomic_set(dst, value, pe):
    """
    Atomically sets the value at symmetric ``dst`` on PE ``pe`` to ``value``.

    This is a thread-level remote atomic store. The write is performed atomically
    with respect to other atomic operations on the same location.

    Args:
        - ``dst``: CuTe tensor view pointing to a single-element symmetric destination on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. The element dtype determines
          which underlying NVSHMEM atomic is dispatched.
        - ``value``: The value to store. Cast to the element dtype of ``dst`` before the operation.
        - ``pe`` (``int``): Target PE.

    Note:
        This operation does not return the old value. Use ``atomic_fetch`` before setting
        if you need the previous value.
        Supported dtypes are determined by the NVSHMEM atomic set dispatch table.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    dst_ptr = _resolve_ptr(dst)
    dtype = _resolve_dtype(dst)
    value = cute_cast(value, dtype)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_set(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_set(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_set(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_set(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Float32):
        return float_atomic_set(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Float64):
        return double_atomic_set(dst_ptr, value, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_set: {dtype}")


@cute.jit
def atomic_compare_swap(dst, cond, value, pe):
    """
    Atomically compares the value at symmetric ``dst`` on PE ``pe`` with ``cond``,
    and if equal, replaces it with ``value``. Returns the old value regardless.

    This is a thread-level remote atomic compare-and-swap (CAS) operation.

    Args:
        - ``dst``: CuTe tensor view pointing to a single-element symmetric destination on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. The element dtype determines
          which underlying NVSHMEM atomic is dispatched.
        - ``cond``: Comparison value. Cast to the element dtype of ``dst``.
          The swap only occurs if the current value at ``dst`` equals ``cond``.
        - ``value``: Replacement value. Cast to the element dtype of ``dst``.
          Written to ``dst`` only if the comparison succeeds.
        - ``pe`` (``int``): Target PE.

    Returns:
        The value stored at ``dst`` on PE ``pe`` prior to the operation, regardless of
        whether the swap occurred.

    Note:
        Supported dtypes are determined by the NVSHMEM atomic compare-swap dispatch table.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    dst_ptr = _resolve_ptr(dst)
    dtype = _resolve_dtype(dst)
    cond = cute_cast(cond, dtype)
    value = cute_cast(value, dtype)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_compare_swap(dst_ptr, cond, value, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_compare_swap(dst_ptr, cond, value, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_compare_swap(dst_ptr, cond, value, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_compare_swap(dst_ptr, cond, value, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_compare_swap: {dtype}")


@cute.jit
def atomic_swap(dst, value, pe):
    """
    Atomically replaces the value at symmetric ``dst`` on PE ``pe`` with ``value``,
    and returns the old value.

    This is a thread-level remote atomic swap operation.

    Args:
        - ``dst``: CuTe tensor view pointing to a single-element symmetric destination on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. The element dtype determines
          which underlying NVSHMEM atomic is dispatched.
        - ``value``: New value to store. Cast to the element dtype of ``dst``.
        - ``pe`` (``int``): Target PE.

    Returns:
        The value stored at ``dst`` on PE ``pe`` prior to the swap.

    Note:
        Supported dtypes are determined by the NVSHMEM atomic swap dispatch table.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    dst_ptr = _resolve_ptr(dst)
    dtype = _resolve_dtype(dst)
    value = cute_cast(value, dtype)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_swap(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_swap(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_swap(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_swap(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Float32):
        return float_atomic_swap(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Float64):
        return double_atomic_swap(dst_ptr, value, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_swap: {dtype}")


@cute.jit
def atomic_fetch_inc(dst, pe):
    """
    Atomically increments the value at symmetric ``dst`` on PE ``pe`` by 1,
    and returns the value prior to the increment.

    This is a thread-level remote atomic fetch-and-increment operation.

    Args:
        - ``dst``: CuTe tensor view pointing to a single-element symmetric destination on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. Supported dtypes are integral types
          as determined by the NVSHMEM atomic fetch-inc dispatch table.
        - ``pe`` (``int``): Target PE.

    Returns:
        The value stored at ``dst`` on PE ``pe`` prior to the increment.

    Note:
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    dst_ptr = _resolve_ptr(dst)
    dtype = _resolve_dtype(dst)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_fetch_inc(dst_ptr, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_fetch_inc(dst_ptr, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_fetch_inc(dst_ptr, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_fetch_inc(dst_ptr, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_fetch_inc: {dtype}")


@cute.jit
def atomic_inc(dst, pe):
    """
    Atomically increments the value at symmetric ``dst`` on PE ``pe`` by 1.
    Does not return the old value.

    This is a thread-level remote atomic increment operation (non-fetching variant).
    Use ``atomic_fetch_inc`` if you need the value before the increment.

    Args:
        - ``dst``: CuTe tensor view pointing to a single-element symmetric destination on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. Supported dtypes are integral types
          as determined by the NVSHMEM atomic inc dispatch table.
        - ``pe`` (``int``): Target PE.

    Note:
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    dst_ptr = _resolve_ptr(dst)
    dtype = _resolve_dtype(dst)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_inc(dst_ptr, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_inc(dst_ptr, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_inc(dst_ptr, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_inc(dst_ptr, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_inc: {dtype}")


@cute.jit
def atomic_fetch_add(dst, value, pe):
    """
    Atomically adds ``value`` to the value at symmetric ``dst`` on PE ``pe``,
    and returns the value prior to the addition.

    This is a thread-level remote atomic fetch-and-add operation.

    Args:
        - ``dst``: CuTe tensor view pointing to a single-element symmetric destination on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. The element dtype determines
          which underlying NVSHMEM atomic is dispatched (integral and floating-point types supported).
        - ``value``: Value to add. Cast to the element dtype of ``dst``.
        - ``pe`` (``int``): Target PE.

    Returns:
        The value stored at ``dst`` on PE ``pe`` prior to the addition.

    Note:
        Supported dtypes are determined by the NVSHMEM atomic fetch-add dispatch table.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    dst_ptr = _resolve_ptr(dst)
    dtype = _resolve_dtype(dst)
    value = cute_cast(value, dtype)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_fetch_add(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_fetch_add(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_fetch_add(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_fetch_add(dst_ptr, value, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_fetch_add: {dtype}")


@cute.jit
def atomic_add(dst, value, pe):
    """
    Atomically adds ``value`` to the value at symmetric ``dst`` on PE ``pe``.
    Does not return the old value.

    This is a thread-level remote atomic add operation (non-fetching variant).
    Use ``atomic_fetch_add`` if you need the value before the addition.

    Args:
        - ``dst``: CuTe tensor view pointing to a single-element symmetric destination on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. The element dtype determines
          which underlying NVSHMEM atomic is dispatched (integral and floating-point types supported).
        - ``value``: Value to add. Cast to the element dtype of ``dst``.
        - ``pe`` (``int``): Target PE.

    Note:
        Supported dtypes are determined by the NVSHMEM atomic add dispatch table.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    dst_ptr = _resolve_ptr(dst)
    dtype = _resolve_dtype(dst)
    value = cute_cast(value, dtype)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_add(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_add(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_add(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_add(dst_ptr, value, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_add: {dtype}")


@cute.jit
def atomic_and(dst, value, pe):
    """
    Atomically applies bitwise AND of ``value`` with the value at symmetric ``dst`` on PE ``pe``.
    Does not return the old value.

    This is a thread-level remote atomic bitwise AND operation (non-fetching variant).
    Use ``atomic_fetch_and`` if you need the value before the operation.

    Args:
        - ``dst``: CuTe tensor view pointing to a single-element symmetric destination on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. Only integral (bitwise) dtypes are supported.
        - ``value``: Mask value to AND with. Cast to the element dtype of ``dst``.
        - ``pe`` (``int``): Target PE.

    Note:
        Only integral dtypes (e.g., ``uint32``, ``uint64``, etc.) are supported.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    dst_ptr = _resolve_ptr(dst)
    dtype = _resolve_dtype(dst)
    value = cute_cast(value, dtype)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_and(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_and(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_and(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_and(dst_ptr, value, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_and: {dtype}")


@cute.jit
def atomic_fetch_and(dst, value, pe):
    """
    Atomically applies bitwise AND of ``value`` with the value at symmetric ``dst`` on PE ``pe``,
    and returns the value prior to the operation.

    This is a thread-level remote atomic fetch-and-AND operation.

    Args:
        - ``dst``: CuTe tensor view pointing to a single-element symmetric destination on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. Only integral (bitwise) dtypes are supported.
        - ``value``: Mask value to AND with. Cast to the element dtype of ``dst``.
        - ``pe`` (``int``): Target PE.

    Returns:
        The value stored at ``dst`` on PE ``pe`` prior to the AND operation.

    Note:
        Only integral dtypes (e.g., ``uint32``, ``uint64``, etc.) are supported.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    dst_ptr = _resolve_ptr(dst)
    dtype = _resolve_dtype(dst)
    value = cute_cast(value, dtype)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_fetch_and(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_fetch_and(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_fetch_and(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_fetch_and(dst_ptr, value, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_fetch_and: {dtype}")


@cute.jit
def atomic_or(dst, value, pe):
    """
    Atomically applies bitwise OR of ``value`` with the value at symmetric ``dst`` on PE ``pe``.
    Does not return the old value.

    This is a thread-level remote atomic bitwise OR operation (non-fetching variant).
    Use ``atomic_fetch_or`` if you need the value before the operation.

    Args:
        - ``dst``: CuTe tensor view pointing to a single-element symmetric destination on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. Only integral (bitwise) dtypes are supported.
        - ``value``: Mask value to OR with. Cast to the element dtype of ``dst``.
        - ``pe`` (``int``): Target PE.

    Note:
        Only integral dtypes (e.g., ``uint32``, ``uint64``, etc.) are supported.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    dst_ptr = _resolve_ptr(dst)
    dtype = _resolve_dtype(dst)
    value = cute_cast(value, dtype)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_or(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_or(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_or(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_or(dst_ptr, value, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_or: {dtype}")


@cute.jit
def atomic_fetch_or(dst, value, pe):
    """
    Atomically applies bitwise OR of ``value`` with the value at symmetric ``dst`` on PE ``pe``,
    and returns the value prior to the operation.

    This is a thread-level remote atomic fetch-and-OR operation.

    Args:
        - ``dst``: CuTe tensor view pointing to a single-element symmetric destination on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. Only integral (bitwise) dtypes are supported.
        - ``value``: Mask value to OR with. Cast to the element dtype of ``dst``.
        - ``pe`` (``int``): Target PE.

    Returns:
        The value stored at ``dst`` on PE ``pe`` prior to the OR operation.

    Note:
        Only integral dtypes (e.g., ``uint32``, ``uint64``, etc.) are supported.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    dst_ptr = _resolve_ptr(dst)
    dtype = _resolve_dtype(dst)
    value = cute_cast(value, dtype)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_fetch_or(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_fetch_or(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_fetch_or(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_fetch_or(dst_ptr, value, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_fetch_or: {dtype}")


@cute.jit
def atomic_xor(dst, value, pe):
    """
    Atomically applies bitwise XOR of ``value`` with the value at symmetric ``dst`` on PE ``pe``.
    Does not return the old value.

    This is a thread-level remote atomic bitwise XOR operation (non-fetching variant).
    Use ``atomic_fetch_xor`` if you need the value before the operation.

    Args:
        - ``dst``: CuTe tensor view pointing to a single-element symmetric destination on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. Only integral (bitwise) dtypes are supported.
        - ``value``: Mask value to XOR with. Cast to the element dtype of ``dst``.
        - ``pe`` (``int``): Target PE.

    Note:
        Only integral dtypes (e.g., ``uint32``, ``uint64``, etc.) are supported.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    dst_ptr = _resolve_ptr(dst)
    dtype = _resolve_dtype(dst)
    value = cute_cast(value, dtype)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_xor(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_xor(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_xor(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_xor(dst_ptr, value, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_xor: {dtype}")


@cute.jit
def atomic_fetch_xor(dst, value, pe):
    """
    Atomically applies bitwise XOR of ``value`` with the value at symmetric ``dst`` on PE ``pe``,
    and returns the value prior to the operation.

    This is a thread-level remote atomic fetch-and-XOR operation.

    Args:
        - ``dst``: CuTe tensor view pointing to a single-element symmetric destination on PE ``pe``.
          Must be a symmetric (NVSHMEM-allocated) tensor. Only integral (bitwise) dtypes are supported.
        - ``value``: Mask value to XOR with. Cast to the element dtype of ``dst``.
        - ``pe`` (``int``): Target PE.

    Returns:
        The value stored at ``dst`` on PE ``pe`` prior to the XOR operation.

    Note:
        Only integral dtypes (e.g., ``uint32``, ``uint64``, etc.) are supported.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    dst_ptr = _resolve_ptr(dst)
    dtype = _resolve_dtype(dst)
    value = cute_cast(value, dtype)


    if const_expr(dtype == cutlass.Int32):
        return int32_atomic_fetch_xor(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Int64):
        return long_atomic_fetch_xor(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint32):
        return uint32_atomic_fetch_xor(dst_ptr, value, pe)


    elif const_expr(dtype == cutlass.Uint64):
        return uint64_atomic_fetch_xor(dst_ptr, value, pe)


    raise RuntimeError(f"Unsupported CuTe dtype for atomic_fetch_xor: {dtype}")


# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import nvshmem.bindings.device.numba as bindings
from nvshmem.core import Teams

import cffi
from numba.core.extending import overload
from numba.types import int8, int16, int32, int64, uint8, uint16, uint32, uint64, float32, float64, float16, Array
from numba import cuda
from numba.np.numpy_support import carray
from numba.np import arrayobj
from numba.cuda.extending import intrinsic

__all__ = ["my_pe", "team_my_pe", "team_n_pes", "n_pes", "barrier_all", "sync_all", "signal_op", "signal_wait"]

# TODO: create a global ffi object for other high level bindings to use
ffi = cffi.FFI()


def my_pe():
    pass


@overload(my_pe)
def my_pe_ol():

    def impl():
        return bindings.my_pe()

    return impl


def team_my_pe():
    pass


@overload(team_my_pe)
def team_my_pe_ol(team):

    def impl(team):
        return bindings.team_my_pe(team)

    return impl


def team_n_pes():
    pass


@overload(team_n_pes)
def team_n_pes_ol(team):

    def impl(team):
        return bindings.team_n_pes(team)

    return impl


def n_pes():
    pass


@overload(n_pes)
def n_pes_ol():

    def impl():
        return bindings.n_pes()

    return impl


def barrier_all():
    pass


@overload(barrier_all)
def barrier_all_ol():

    def impl():
        return bindings.barrier_all()

    return impl


def sync_all():
    pass


@overload(sync_all)
def sync_all_ol():

    def impl():
        return bindings.sync_all()

    return impl


def signal_op():
    pass


@overload(signal_op)
def signal_op_ol(signal_var, signal_val, signal_op, pe):

    def impl(signal_var, signal_val, signal_op, pe):
        signal_varptr = ffi.from_buffer(signal_var)
        bindings.signal_op(signal_varptr, signal_val, signal_op, pe)

    return impl


def signal_wait():
    pass


@overload(signal_wait)
def signal_wait_ol(signal_var, signal_op, signal_val):

    def impl(signal_var, signal_op, signal_val):
        signal_varptr = ffi.from_buffer(signal_var)
        return bindings.signal_wait_until(signal_varptr, signal_op, signal_val)

    return impl



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
Internal tracking for NVShmem

This file contains things like buffer management, status, etc.
"""
from enum import IntEnum
"""
Map of Device IDs from cuda.core to MemoryResource (NvshmemResource) objects
Used to avoid re-creating NvshmemResources every time someone calls nvshmem.core.allocate()
"""
_mr_references = {}
"""
class for Internal Init Status
"""


class InternalInitStatus(IntEnum):
    UNINITIALIZED = 0
    INITIALIZED = 1
    DE_INITIALIZED = 2  # Keeps bootstrap


"""
Set to True after initializing. Used for safety checks before functions
"""
_is_initialized = {"status": InternalInitStatus.UNINITIALIZED}
"""
Each NVSHMEM process needs to be assocaited with a device. We cache that here.
"""
_cached_device = {"device": None}
"""
Debug mode is used to avoid redundant calls to Device()
"""
_debug_mode = False



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from .nvshmem import *

# Define what gets exposed when users do `import nvshmem.bindings`
__all__ = [name for name in dir() if not name.startswith("_")]



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
These functions are NVSHMEM4Py APIs that expose host-initiated remote memory accesses (RMA)
"""

from cuda.core import Stream
from cuda.core import Buffer

from nvshmem.core.interop.cupy import _is_array, array_get_buffer
from nvshmem.core.interop.torch import _is_tensor, tensor_get_buffer
from nvshmem.core._internal_tracking import _is_initialized, InternalInitStatus
from nvshmem.core.nvshmem_types import *
from nvshmem.core.utils import _get_device
import nvshmem.bindings as bindings
from nvshmem.core.direct import ComparisonType, SignalOp

import logging
from enum import IntEnum
from typing import Tuple

__all__ = ["put_signal", "signal_op", "signal_wait", "put", "get", "quiet", "SignalOp"]

logger = logging.getLogger("nvshmem")


def _get_buffers(dst, src) -> Tuple[Buffer, Buffer]:
    """
    Converts high-level inputs into cuda.core Buffer objects.

    Supports inputs as Cupy arrays, Torch tensors, or Buffer instances.
    Both ``dst`` and ``src`` must resolve to valid Buffers for the operation to proceed.

    Args:
        - dst (object): Destination memory object.
        - src (object): Source memory object.

    Returns:
        ``Tuple[Buffer, Buffer]``: A tuple of (dst_buffer, src_buffer).

    Raises:
        ``NvshmemInvalid``: If either input cannot be converted into a valid Buffer.
    """
    src_buf = None
    dst_buf = None

    if _is_array(src):
        src_buf, size, type = array_get_buffer(src)
    if _is_array(dst):
        dst_buf, size, type = array_get_buffer(dst)

    if _is_tensor(src):
        src_buf, size, type = tensor_get_buffer(src)
    if _is_tensor(dst):
        dst_buf, size, type = tensor_get_buffer(dst)

    if isinstance(dst, Buffer):
        dst_buf = dst
    if isinstance(src, Buffer):
        src_buf = src

    # if either are None at this point, it's invalid
    if src_buf is None or dst_buf is None:
        raise NvshmemInvalid("Invalid memory objects passed to NVSHMEM operation")

    return dst_buf, src_buf


def _call_putget(dst: object,
                 src: object,
                 op: str = "put",
                 signal: bool = False,
                 signal_var: Buffer = 0,
                 signal_val: int = 0,
                 signal_op: SignalOp = 0,
                 remote_pe: int = 0,
                 stream: NvshmemStreamsType = None) -> None:
    """
    Internal helper to invoke host-initiated NVSHMEM put/get with optional signaling.

    This function wraps low-level C bindings and computes the safe minimum transfer size.
    If `signal` is enabled, the provided `signal_var` must be a symmetric NVSHMEM buffer
    of at least 8 bytes.

    Args:
        - dst (object): Destination buffer (Buffer, Cupy array, or Torch tensor).
        - src (object): Source buffer (Buffer, Cupy array, or Torch tensor).
        - op (str): Either "put" or "get".
        - signal (bool): If True, invoke signal-enabled variant of the op.
        - signal_var (Buffer): Symmetric memory buffer used as the signal address.
                             This buffer must be >=8 bytes in size (the underlying API expects a uint64)
        - signal_val (int): Value to use in the signal operation.
        - signal_op (SignalOp): Signal operation type.
        - remote_pe (int): Target PE for the remote memory access.
        - stream (Stream): CUDA stream to issue the operation on.

    Raises:
        ``NotImplementedError``: If stream is None (non-stream variants unsupported).
        ``ValueError``: If op is not "put" or "get", or if inputs are invalid.
        ``NvshmemError``: If any operations do not complete successfully
    """
    if _is_initialized["status"] != InternalInitStatus.INITIALIZED:
        raise NvshmemInvalid("NVSHMEM Library is not initialized")

    user_nvshmem_dev, other_dev = _get_device()
    dst_buf, src_buf = _get_buffers(dst, src)
    if stream is None:
        logger.error("Non on-stream put/get operations are not yet implemented")
        raise NotImplemented

    if op not in ("put", "get"):
        raise NvshmemInvalid("Tried to call put/get function with an operation not put nor get")

    if not isinstance(src_buf, Buffer) or not isinstance(dst_buf, Buffer):
        raise NvshmemInvalid("Called collective on an invalid Buffer")

    f_name = f"{op}mem{'_signal' if signal else ''}_on_stream"
    func = getattr(bindings, f_name)
    safe_size = min(src_buf.size, dst_buf.size)
    if signal:
        if not isinstance(signal_var, Buffer) or signal_var.size < 8:
            raise NvshmemInvalid("Signal must be a Buffer >= 8 bytes allocated by NVSHMEM4Py")
        f_args = [
            dst_buf.handle, src_buf.handle, safe_size, signal_var.handle, signal_val, signal_op, remote_pe,
            int(stream.__cuda_stream__()[1])
        ]
    else:
        f_args = [dst_buf.handle, src_buf.handle, safe_size, remote_pe, int(stream.__cuda_stream__()[1])]

    func(*f_args)

    if other_dev is not None:
        other_dev.set_current()


def put_signal(dst: object,
               src: object,
               signal_var: Buffer,
               signal_val: int,
               signal_op: SignalOp,
               remote_pe: int = -1,
               stream=None) -> None:
    """
    Performs a put with signal on a CUDA stream.

    Args:
        - dst (object): Destination buffer (Buffer, Cupy array, or Torch tensor).
        - src (object): Source buffer (Buffer, Cupy array, or Torch tensor).
        - signal_var (Buffer): Symmetric memory buffer used as signal variable.
        - signal_val (int): Value to use in the signal operation.
        - signal_op (SignalOp): Signal operation type.
        - remote_pe (int): Target PE for the put.
        - stream (Stream): CUDA stream to issue the put on.

    Raises:
        - ``NotImplementedError``: If stream is None.
        - ``ValueError``: If the signal buffer is invalid or too small.
        - ``NvshmemError``: If any operations do not complete successfully
    """
    _call_putget(dst,
                 src,
                 op="put",
                 signal=True,
                 signal_var=signal_var,
                 signal_val=signal_val,
                 signal_op=signal_op,
                 remote_pe=remote_pe,
                 stream=stream)


def signal_op(signal_var: Buffer,
              signal_val: int,
              signal_op: SignalOp,
              remote_pe: int = -1,
              stream: NvshmemStreamsType = None) -> None:
    """
    Performs a signal operation on a CUDA stream. Equivalent to a ``put_signal`` but with no data movement.

    Args:
        - signal_var (Buffer): Symmetric memory buffer used as the signal variable.
        - signal_val (int): Value to use in the signal operation.
        - signal_op (SignalOp): Signal operation type.
        - remote_pe (int): Target PE for the signal operation.
        - stream (Stream): CUDA stream to issue the signal operation on.

    Raises:
        - ``NotImplementedError``: If stream is None.
        - ``ValueError``: If the signal buffer is invalid or too small.
        - ``NvshmemError``: If any operations do not complete successfully
    """
    if _is_initialized["status"] != InternalInitStatus.INITIALIZED:
        raise NvshmemInvalid("NVSHMEM Library is not initialized")
    user_nvshmem_dev, other_dev = _get_device()
    if not isinstance(signal_var, Buffer) or signal_var.size < 8:
        raise NvshmemInvalid("Signal must be a Buffer >= 8 bytes allocated by NVSHMEM4Py")
    if stream is None:
        logger.error("Non on-stream signal operations are not yet implemented")
        raise NotImplemented
    bindings.signal_op_on_stream(signal_var.handle, signal_val, signal_op, remote_pe, int(stream.__cuda_stream__()[1]))
    if other_dev is not None:
        other_dev.set_current()


def signal_wait(signal_var: Buffer,
                signal_val: int,
                signal_op: ComparisonType,
                stream: NvshmemStreamsType = None) -> None:
    """
    Waits until a symmetric signal variable satisfies a given condition.

    Args:
        - signal_var (Buffer): Symmetric memory buffer used as the signal source.
        - signal_val (int): Value to compare against.
        - signal_op (SignalOp): Wait condition
        - stream (Stream): CUDA stream to issue the wait on.

    Raises:
        - ``NotImplementedError``: If stream is None.
        - ``ValueError``: If the signal buffer is invalid.
        - ``NvshmemError``: If any operations do not complete successfully
    """
    if stream is None:
        logger.error("Non on-stream put/get operations are not yet implemented")
        raise NotImplemented
    user_nvshmem_dev, other_dev = _get_device()

    bindings.signal_wait_until_on_stream(signal_var.handle, signal_op, signal_val, int(stream.__cuda_stream__()[1]))

    if other_dev is not None:
        other_dev.set_current()


def quiet(stream: NvshmemStreamsType = None) -> None:
    """
    Ensures completion of all previously issued NVSHMEM operations on the given stream.

    This is equivalent to a device-side ``nvshmem_quiet`` for host-initiated NVSHMEM operations.

    Args:
        - stream (``Stream``): CUDA stream to synchronize.

    Raises:
        - ``NotImplementedError``: If stream is None.
        - ``NvshmemError``: If any operations do not complete successfully
    """
    if stream is None:
        logger.error("Non on-stream put/get operations are not yet implemented")
        raise NotImplemented
    user_nvshmem_dev, other_dev = _get_device()
    # Because quiet doesn't have a datatype, it's a special case and doesn't need to use _call_putget function
    bindings.quiet_on_stream(int(stream.__cuda_stream__()[1]))
    if other_dev is not None:
        other_dev.set_current()


def put(dst: object, src: object, remote_pe: int = -1, stream: NvshmemStreamsType = None):
    """
    Performs a host-initiated NVSHMEM put operation on a CUDA stream.

    Args:
        - dst (object): Destination buffer (Buffer, Cupy array, or Torch tensor).
        - src (object): Source buffer (Buffer, Cupy array, or Torch tensor).
        - remote_pe (int): Target PE for the put.
        - stream (Stream): CUDA stream to issue the put on.

    Raises:
        - ``NotImplementedError``: If stream is ``None``.
        - ``NvshmemInvalid``: If inputs are not valid Buffer-compatible types.
        - ``NvshmemError``: If any operations do not complete successfully
    """
    _call_putget(dst, src, op="put", signal=False, remote_pe=remote_pe, stream=stream)


def get(dst: object, src: object, remote_pe: int = -1, stream: NvshmemStreamsType = None):
    """
    Performs a host-initiated NVSHMEM get operation on a CUDA stream.

    Args:
        - dst (object): Destination buffer (Buffer, Cupy array, or Torch tensor).
        - src (object): Source buffer (Buffer, Cupy array, or Torch tensor).
        - remote_pe (int): Target PE for the get.
        - stream (Stream): CUDA stream to issue the get on.

    Raises:
        - ``NotImplementedError``: If stream is None.
        - ``ValueError``: If inputs are not valid Buffer-compatible types.
        - ``NvshmemError``: If any operations do not complete successfully
    """
    _call_putget(dst, src, op="get", signal=False, remote_pe=remote_pe, stream=stream)
